import numpy as np
import queue
import copy


def cost_function(gallery, sensor_distribution):
    """
    Method to calculate the cost of the cells in the grid given the current sensor distribution

    Parameters
    ----------
    gallery: Gallery
        Map of the environment with information about sensors of given size
    sensor_distribution: list(list(double), list(int))
        List of two lists. In the second list are the sensors, and the first list is the corresponding sensor's probability.
    Returns
    -------
    cost: Matrix of the same size as the map in the gallery
        Matrix filled by cost for each grid cell in the environment. (Costs are set in constants FREE_COST and SENSOR_COST.
        The cost for accessing a cell is FREE_COST + SENSOR_COST * probability of the sensor being active in the cell.)
    """

    return np.full((gallery.y_size, gallery.x_size), 0.01)


def best_plan(gallery, cost_matrix):
    """
    Method to calculate the best path to the goal and back given the current cost function

    Parameters
    ----------
    gallery: Gallery
        Map of the environment with information about sensors of a given size
    cost_matrix: Matrix of the same size as the map in the gallery
        Matrix capturing the cost of each cell
    Returns
    -------
    path: list(tuple(int, int))
        List of coordinates visited on the best path
    value: double
        Value of the best path given the cost matrix
    """
    q = queue.Queue()
    q.put((gallery.entrances[0], [gallery.entrances[0]]))

    closed = set()

    while True:
        cell, path = q.get()
        if cell in closed:
            continue
        if cell in gallery.goals:
            break
        closed.add(cell)
        for neighbour in passable_neighbours(cell, gallery):
            new_path = copy.copy(path)
            new_path.append(neighbour)
            q.put((neighbour, new_path))

    return path + list(reversed(path[:-1])), 0


def passable_neighbours(cell, gallery):
    neighbours = []
    for move in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
        possible_neighbour = (cell[0] + move[0], cell[1] + move[1])
        if gallery.is_correct_cell(possible_neighbour) and gallery.is_passable(possible_neighbour):
            neighbours.append(possible_neighbour)
    return neighbours
